/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.onlineSim.cac;

import cern.jet.random.tdouble.AbstractDoubleDistribution;
import cern.jet.random.tdouble.Exponential;
import cern.jet.random.tdouble.engine.MersenneTwister64;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.interfaces.simulation.IEventGenerator;
import com.net2plan.interfaces.simulation.SimEvent;
import com.net2plan.utils.Pair;
import com.net2plan.utils.RandomUtils;
import com.net2plan.utils.StringUtils;
import com.net2plan.utils.Triple;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

/**
 * Connection generator in which connection requests from traffic demands arrive
 * according to a Poisson process and are independent of each other.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class CACSim_EG_exponentialConnectionGenerator extends IEventGenerator
{
	private boolean incrementalModel;
	private long trafficLayerId, numConnectionsSimulation, numConnectionsTransitory, currentNumConnections;
	private Map<Long, Pair<Long, Long>> originalDemandMap;
	private Map<Long, AbstractDoubleDistribution> iat_d;
	private Map<Long, AbstractDoubleDistribution> ht_d;
	private Map<Long, Double> s_d;
	
	@Override
	public String getDescription()
	{
		return "Connection generator in which connection requests from traffic "
				+ "demands arrive according to a Poisson process and are "
				+ "independent of each other";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> parameters = new LinkedList<Triple<String, String, String>>();
		parameters.add(Triple.of("defaultConnectionSize", "1", "Default requested traffic volume per connection"));
		parameters.add(Triple.of("defaultHoldingTime", "120", "Default average connection duration (in seconds)"));
		parameters.add(Triple.of("numConnectionsSimulation", "-1", "Number of connections to finish the simulation (a negative or zero value means no limit)"));
		parameters.add(Triple.of("numConnectionsTransitory", "-1", "Number of connections to finish the transitory (a negative or zero value means no transitory)"));
		parameters.add(Triple.of("randomSeed", "-1", "Seed for the random generator (-1 means random)"));
		parameters.add(Triple.of("trafficLayerId", "-1", "Layer containing traffic demands (-1 means default layer)"));
		parameters.add(Triple.of("simulationModel", "#select# longRun incrementalModel", "Simulation model: 'longRun' (connections are established and released), 'incrementalModel' (connections are never released)"));

		return parameters;
	}

	@Override
	public void initialize(NetPlan initialNetPlan, NetPlan currentNetPlan, Map<String, String> algorithmParameters, Map<String, String> simulationParameters, Map<String, String> net2planParameters)
	{
		trafficLayerId = Long.parseLong(algorithmParameters.get("trafficLayerId"));
		if (trafficLayerId == -1) trafficLayerId = currentNetPlan.getLayerDefaultId();
		
		int D = initialNetPlan.getNumberOfDemands(trafficLayerId);
		if (D == 0) throw new Net2PlanException("No demands were defined in the original design");

		long seed = Long.parseLong(algorithmParameters.get("randomSeed"));
		if (seed == -1)	seed = RandomUtils.random(0, Long.MAX_VALUE - 1);
		Random seedGenerator = new Random(seed);

		int defaultConnectionSize = Integer.parseInt(algorithmParameters.get("defaultConnectionSize"));
		if (defaultConnectionSize <= 0) throw new Net2PlanException("'defaultConnectionSize' must be greater than zero");

		double defaultHoldingTime = Double.parseDouble(algorithmParameters.get("defaultHoldingTime"));
		if (defaultHoldingTime <= 0) throw new Net2PlanException("'defaultHoldingTime' must be greater than zero");
		
		numConnectionsSimulation = Long.parseLong(algorithmParameters.get("numConnectionsSimulation"));
		numConnectionsTransitory = Long.parseLong(algorithmParameters.get("numConnectionsTransitory"));
		if (numConnectionsSimulation <= 0) numConnectionsSimulation = Long.MAX_VALUE;
		if (numConnectionsTransitory <= 0) numConnectionsTransitory = Long.MAX_VALUE;
		
		String simulationModel = algorithmParameters.get("simulationModel");
		if (!simulationModel.equalsIgnoreCase("longRun") && !simulationModel.equalsIgnoreCase("incrementalModel")) throw new Net2PlanException("Simulation model must be either 'longRun' or 'incrementalModel'");
		incrementalModel = simulationModel.equalsIgnoreCase("incrementalModel");

		currentNetPlan.removeAllDemands(trafficLayerId);
		originalDemandMap = initialNetPlan.getDemandMap(trafficLayerId);
		iat_d = new LinkedHashMap<Long, AbstractDoubleDistribution>();
		ht_d = new LinkedHashMap<Long, AbstractDoubleDistribution>();
		s_d = new LinkedHashMap<Long, Double>();

		Set<Long> originalDemandIds = initialNetPlan.getDemandIds(trafficLayerId);
		for (long originalDemandId : originalDemandIds)
		{
			double h_d = initialNetPlan.getDemandOfferedTraffic(trafficLayerId, originalDemandId);

			String aux_str_s_d = initialNetPlan.getDemandAttribute(trafficLayerId, originalDemandId, "connectionSize");
			double aux_s_d;
			try	{ aux_s_d = Double.parseDouble(aux_str_s_d); }
			catch (Throwable e) { aux_s_d = defaultConnectionSize; }
			if (aux_s_d <= 0) throw new Net2PlanException("'connectionSize' for demand " + originalDemandId + " is lower or equal than zero");

			String aux_str_holdTime = initialNetPlan.getDemandAttribute(trafficLayerId, originalDemandId, "holdingTime");
			double aux_holdTime;
			try { aux_holdTime = Double.parseDouble(aux_str_holdTime); }
			catch (Throwable e) { aux_holdTime = defaultHoldingTime; }
			if (aux_holdTime <= 0) throw new Net2PlanException("'holdingTime' for demand " + originalDemandId + " is lower or equal than zero");

			s_d.put(originalDemandId, aux_s_d);
			double holdingTime = aux_holdTime;
			double interArrivalTime = s_d.get(originalDemandId) * holdingTime / h_d;

			iat_d.put(originalDemandId, new Exponential(1 / interArrivalTime, new MersenneTwister64(seedGenerator.nextInt())));
			ht_d.put(originalDemandId, new Exponential(1 / holdingTime, new MersenneTwister64(seedGenerator.nextInt())));

			scheduleNewConnectionArrival(originalDemandId, 0);
		}

		currentNumConnections = 0;
	}

	@Override
	public void processEvent(NetPlan currentNetPlan, SimEvent event)
	{
		double simTime = event.getEventTime();
		Object eventObject = event.getEventObject();
		if (eventObject instanceof ConnectionRequest)
		{
			if (numConnectionsTransitory != -1 && currentNumConnections >= numConnectionsTransitory)
			{
				endTransitory();
				numConnectionsTransitory = -1;
			}
			
			if (numConnectionsSimulation != -1 && currentNumConnections >= numConnectionsSimulation)
			{
				endSimulation();
			}
			
			ConnectionRequest connectionRequest = (ConnectionRequest) eventObject;
			long originalDemandId = connectionRequest.originalDemandId;
			double duration = connectionRequest.duration;

			Pair<Long, Long> ingressEgressNodePair = originalDemandMap.get(originalDemandId);
			long ingressNodeId = ingressEgressNodePair.getFirst();
			long egressNodeId = ingressEgressNodePair.getSecond();
			
			currentNetPlan.activateActionMonitoring();

			Map<String, String> demandAttributes = new LinkedHashMap<String, String>();
			demandAttributes.put("arrivalTime", StringUtils.secondsToYearsDaysHoursMinutesSeconds(simTime));
			if (!incrementalModel) demandAttributes.put("duration", StringUtils.secondsToYearsDaysHoursMinutesSeconds(duration));
			if (!incrementalModel) demandAttributes.put("departureTime", StringUtils.secondsToYearsDaysHoursMinutesSeconds(simTime + duration));
			demandAttributes.put("originalDemandId", Long.toString(originalDemandId));
			long demandId = currentNetPlan.addDemand(trafficLayerId, ingressNodeId, egressNodeId, s_d.get(originalDemandId), demandAttributes);

			scheduleEvent(new SimEvent(simTime, currentNetPlan.getActionsPerformed(), SimEvent.DestinationModule.EVENT_PROCESSOR));
			if (!incrementalModel) scheduleEvent(new SimEvent(simTime + duration, new ConnectionRelease(demandId), SimEvent.DestinationModule.EVENT_GENERATOR));
			scheduleNewConnectionArrival(originalDemandId, simTime);

			currentNumConnections++;
		}
		else if (eventObject instanceof ConnectionRelease)
		{
			if (incrementalModel) throw new Net2PlanException("In incremental model connections are never released");
			ConnectionRelease connectionRelease = (ConnectionRelease) eventObject;
			long currentDemandId = connectionRelease.currentDemandId;
			if (!currentNetPlan.isDemandActive(trafficLayerId, currentDemandId)) return;
			
			currentNetPlan.activateActionMonitoring();
			currentNetPlan.removeDemand(trafficLayerId, currentDemandId);
			
			scheduleEvent(new SimEvent(simTime, currentNetPlan.getActionsPerformed(), SimEvent.DestinationModule.EVENT_PROCESSOR));
		}
	}

	private void scheduleNewConnectionArrival(long originalDemandId, double simTime)
	{
		double nextArrivalTime = simTime + iat_d.get(originalDemandId).nextDouble();
		double nextDuration = incrementalModel ? Double.MAX_VALUE : ht_d.get(originalDemandId).nextDouble();
		
		scheduleEvent(new SimEvent(nextArrivalTime, new ConnectionRequest(originalDemandId, nextDuration), SimEvent.DestinationModule.EVENT_GENERATOR));
	}

	private static class ConnectionRelease
	{
		public final long currentDemandId;
		
		public ConnectionRelease(long currentDemandId)
		{
			this.currentDemandId = currentDemandId;
		}
		
		@Override
		public String toString() { return "Release connection " + currentDemandId; }
	}

	private static class ConnectionRequest
	{
		public final long originalDemandId;
		public final double duration;
		
		public ConnectionRequest(long originalDemandId, double duration)
		{
			this.originalDemandId = originalDemandId;
			this.duration = duration;
		}
		
		@Override
		public String toString() { return "Connection request for original demand " + originalDemandId + ", duration = " + StringUtils.secondsToYearsDaysHoursMinutesSeconds(duration); }
	}
}